from typing import Dict, List, Optional, Any, Tuple
import freesif #from freesif import open_sif    
import shutil
import numpy as np
import sys
import os
import uuid
import random
import string

def id_generator(size:int=6, chars:str=string.ascii_uppercase + string.digits) -> str:
    """Generates a random string for use as a unique identifier (id).

    Args:
        size (int, optional): Length of random string (id) to be generated. Defaults to 6.
        chars (str, optional): String from which to randomly extract characters for the id. Defaults to string.ascii_uppercase+string.digits.

    Returns:
        str: id
        
    TODO: Move to an external library
    """
    return ''.join(random.choice(chars) for _ in range(size))

def open_sif(folder:str, filename:str)->Tuple[freesif.File, str]:
    """Opens a wadam sif file. This is wrapper to freesif's open_sif function that allows for
    '.' separators in the filename. Should perhaps be extended to include arbitrary extensions.

    Args:
        folder (str): folder containing sif file
        filename (str): name of sif file

    Returns:
        (sif object, temporary filename): Tuple containing the sif object and temporary filename
    """
    id = id_generator()
    filename_tmp = "r" + id + filename[:-4].replace(".","_") + '.SIF'
    shutil.copyfile(f"{folder}/{filename}", filename_tmp)
    # Open the sif and read the data
    sif = freesif.open_sif(filename_tmp)
    return sif, filename_tmp


def get_wadam_strings(folder:str, filename:str, wdfactors:List[float]=[1]*6, 
    pmin:List[float]=[10]*6, pmax:List[float]=[30]*6, ppeak:List[float]=[18.0]*6, 
    scaling:List[str] = ["cos2"]*6, return_arrays:bool=False, 
    return_hz_dirs:bool=False,imosym:int=1,exwave_params:Dict={})->List[List[Any]]:
    """Reads a WADAM SIF file and returns [raoStrings, wavedriftStrings]
    which are lists of length 6, where the 6 entries relate to the degrees of 
    freedom of body motion (surge, sway, heave, roll, pitch, yaw) respectively. 
    
    Each entry is a string which is in SIMO (sys.dat) format, where e.g.
        raoStrings[0] is the surge RAO as a single string
        wavedriftStrings[3] is the mean wavedrift coefficients in roll.
        
    In addition to the strings, the user can get the coefficients as value arrays,
    and can request the frequencies (Hz) and directions (degrees)::
    
        raoStr, wdStr = get_wadam_strings(folder,filename)
        freqs, dirs, raoStr, wdStr = get_wadam_strings(folder,filename,return_hz_dirs=True)
        freqs, dirs, raoStr, wdStr, raos, wds = get_wadam_strings(folder,filename,return_hz_dirs=True, return_arrays=True)      

    Optionally the user can scale the wave drift coefficients using a scaling function.
    The scaling function is specified by the "scaling" argument and can take the values:
    
        "constant": constant scaling
        "linear":   linear scaling. At wave periods <= pmin and >= pmax the scaling factor takes the value of 1.0
                    i.e. no scaling. At wave period ppeak, where pmin < ppeak < pmax, the scaling factor takes a
                    value specified by the corresponding entry in the wdfactors list for the response.
        "cos2" :    Similar to linear except the scaling function between pmin/pmax and ppeak is cosine-squared
        "gauss":    Similar to cos2 except the scaling function is Guassian and the scaling values at pmin/pmin
                    10% the value specified by the wdfactor array
        "exwave":   Uses TR2396 exwave formula. Requires a dictionary of exwave_params containing
                        {"Hs": significant wave height in m,
                         "D0": column diameter,
                         "U":  current speed,
                         "ncols": number of columns}

    Args:
        folder (str): folder of the SIF file
        filename (str): filename of the SIF file
        wdfactors (List[float], optional): Maximum value of scaling function to apply at ppeak. Defaults to [1]*6.
        pmin (List[float], optional): Minimum period for scaling. Defaults to [10]*6.
        pmax (List[float], optional): Maximum period for scaling. Defaults to [30]*6.
        ppeak (List[float], optional): Peak period for scaling. Defaults to [18.0]*6.
        scaling (List[str], optional): Type of scaling. Defaults to ["cos2"]*6.
        return_arrays (bool, optional): Set true to return values of raos and wavedrift coefficients. Defaults to False.
        return_hz_dirs (bool, optional): Set True to return wave frequencies (Hz) and directions. Defaults to False.

    Returns:
        List[List[Any]]: See docstring
    """
    # Read the mean wave drift coefficients and raos on the SIF file
    sif, filename_tmp = open_sif(folder,filename)
    periods = sif.get_periods()
    directions = sif.get_directions()
    omegas = sif.get_angular_freqs()
    raos = sif.get_motion_raos()
    pressure = sif.get_meandrift()
    momentum = sif.get_horiz_meandrift()
    sif.close()
    os.remove(filename_tmp)
    
    # Reformat with increasing frequency    
    raos = raos[:, :, ::-1] # flip the last field
    momentum = momentum[:,:,::-1]
    pressure = pressure[:,:,::-1]
    periods = periods[::-1] # for consistency (superfluous for function)
    omegas = omegas[::-1]
    
    # Compile wavedrift as a mixture of momentum and pressure
    wavedrift = pressure
    wavedrift[0] = momentum[0]
    wavedrift[1] = momentum[1]
    wavedrift[5] = momentum[2]
    wavedrift = wavedrift / 1000. 


    # Establish some blending functions to scale the wavedrifts
    alpha = np.zeros((6,len(periods))) # storage for blending function
    for i in range(0, 6):
        print(i, scaling[i])
        ix_lower = (periods >= pmin[i]) & (periods <= ppeak[i]) 
        ix_upper = (periods <= pmax[i]) & (periods >= ppeak[i])
        coord_lower = 1-(ppeak[i]-periods[ix_lower])/(ppeak[i] - pmin[i])
        coord_upper = 1-(periods[ix_upper]-ppeak[i])/(pmax[i] - ppeak[i])
        if scaling[i] == "constant":
            alpha[i,:] = 1
        if scaling[i] == "linear":
            alpha[i,ix_lower] = coord_lower
            alpha[i,ix_upper] = coord_upper
        if scaling[i] == "cos2":
            alpha[i,ix_lower] = (0.5*(1 - np.cos(coord_lower*np.pi)))**2
            alpha[i,ix_upper] = (0.5*(1 - np.cos(coord_upper*np.pi)))**2
        if scaling[i] == "gauss":
            fwtm_lower = 2*(ppeak[i]-pmin[i]) # full width at tenth of the maximum
            fwtm_upper = 2*(pmax[i]-ppeak[i]) # full width at tenth of the maximum
            c_lower = fwtm_lower/4.29193
            c_upper = fwtm_upper/4.29193
            alpha_lower = np.exp(-(periods-ppeak[i])**2/2/c_lower**2)
            alpha_upper = np.exp(-(periods-ppeak[i])**2/2/c_upper**2)
            alpha[i] = np.where(periods < ppeak[i], alpha_lower, alpha_upper)
        if scaling[i] == "exwave":
            # TODO: requires hs snd other arguments, disallow for now
            hs = exwave_params["Hs"]
            D0 = exwave_params["D0"] #96
            U = exwave_params["U"] #0.73 # would need to enter
            G = 10 # s
            Cp = 0.25 # s/m
            ncols = exwave_params["ncols"]
            Dsum = D0*ncols 
            w = 2*np.pi/periods
            k = w**2/9.80665 # use deep water
            p = np.exp(-1.25*(k*D0)**2)
            B = k*p*Dsum
            exwave_params["scale"] = (1+Cp*U)
            exwave_params["loc"] = B*(G*U +  hs)  

    # if the factor is 1.0 want no change. If factor 1.1 we want some change
    wavedrift_original = wavedrift*1.0 # make a copy for checking (then remove!)
    wavedrift_peak = wavedrift*1.0 # make storage for peak scaling
    for idof in range(len(raos)):
        for idir in range(len(directions)):
            if scaling[idof] == "exwave":
                wavedrift[idof,idir] = wavedrift_original[idof,idir]*exwave_params["scale"] + \
                                            exwave_params["loc"] 
            else:
                wavedrift_peak[idof,idir] = wavedrift_original[idof,idir]*wdfactors[idof]
                wavedrift[idof,idir] = alpha[idof]*wavedrift_peak[idof,idir] + (1-alpha[idof])*wavedrift_original[idof,idir]

    rao_string = []
    rao_string.append("FIRST ORDER MOTION TRANSFER FUNCTION\n\n")
    rao_string.append(f"{directions.size} {omegas.size} {imosym} 2")
    rao_string.append("WAVE DIRECTIONS MOTION TRANSFER FUNCTION")
    rao_string.append("\n".join([f"{idir+1:2d}  {d:.1f}" for idir, d in enumerate(directions)]))
    rao_string.append("WAVE FREQUENCIES MOTION TRANSFER FUNCTION")
    rao_string.append("\n".join([f"{iw+1:2d}  {w:.10e}" for iw, w in enumerate(omegas)]))
    names = "SURGE SWAY HEAVE ROLL PITCH YAW".split()
    for idof, name in enumerate(names):
        rao_string.append(f"{name} MOTION TRANSFER FUNCTION")
        rao_string.append('\n'.join([f"{idir+1:d}  {iw+1:d}  {np.abs(raos[idof,idir,iw]):.10e} {np.angle(raos[idof,idir,iw])*180./np.pi:.10e}" 
                            for idir in range(directions.size) for iw in range(omegas.size)]))
    rao_string = "\n".join(rao_string)

    wd_string = []
    wd_string.append("SECOND ORDER WAVE DRIFT FORCES\n\n")
    wd_string.append(f"{directions.size} {omegas.size} {imosym}")
    wd_string.append("WAVE DIRECTIONS DRIFT COEFFICIENTS")
    wd_string.append("\n".join([f"{idir+1:2d}  {d:.1f}" for idir, d in enumerate(directions)]))
    wd_string.append("WAVE FREQUENCIES DRIFT COEFFICIENTS")
    wd_string.append("\n".join([f"{iw+1:2d}  {w:.10e}" for iw, w in enumerate(omegas)]))
    for idof, name in enumerate(names):
        wd_string.append(f"{name} WAVE DRIFT COEFFICIENTS")
        wd_string.append('\n'.join([f"{idir+1:d}  {iw+1:d}  {wavedrift[idof,idir,iw]:.10e}" 
                            for idir in range(directions.size) for iw in range(omegas.size)]))
    wd_string = "\n".join(wd_string)    

    dataset = [1/periods, directions] if return_hz_dirs else []
    dataset += [rao_string, wd_string]
    dataset += [raos, wavedrift] if return_arrays else []
    return dataset



def read_SIF_rao(folder:str, filename:str, wadir:float, 
                 mapping_index:List[int]=[0,1,2,3,4,5],
                 mapping_sign:List[int]=[1,1,1,1,1,1]):
    """Returns the RAOs on the SIF file for a given wave direction.

    Args:
        folder (str): folder containing the SIF file
        filename (str): filename of the SIF file
        wadir (float): wave direction of interest
        mapping_index (List[int], optional): Maps the degrees of freedom to another order. 
                Defaults to [0,1,2,3,4,5], e.g. [1,0,2,4,3,5] surge<=>sway, roll<=>pitch. 
                Wadir still needs to be correct for unmapped setup. 
        mapping_sign (List[int], optional): Applies a sign change to the mapped setup. 

    Returns:
        [type]: [description]
    """
    
    # Read data
    sif, filename_tmp = open_sif(folder, filename)
    periods = sif.get_periods()
    directions = sif.get_directions()
    raos = sif.get_motion_raos()
    sif.close()
    os.remove(filename_tmp)
    # Extract wave direction and scale.
    ix = np.nonzero(wadir==directions)
    factor = [1,1,1,180./np.pi,180./np.pi,180./np.pi]
    # Correcting for the chosen coordinate system :/
    return 1./periods[::-1], [mapping_sign[dof]*factor[dof]*np.squeeze(raos[dof, ix, ::-1]) for dof in mapping_index]



def get_wave_force_coefficients(folder:str, filename:str, as_simo_string=True, 
                                alpha=[1.0]*6, imosym=1, as_kN=True, blend=[1,1,1,1,1,1],
                                Be:np.ndarray=np.zeros((6,6)), Ae:np.ndarray=np.zeros((6,6)), 
                                Ce:np.ndarray=np.zeros((6,6)))->Tuple[np.ndarray,np.ndarray,np.ndarray]:
    """Returns the wave force coefficients from a WADAM SIF file. If blend=[0,0,0,0,0,0], the function
    will return values on the SIF file. Allows the user to manipulate the returned values by entering 
    part of the external matrices used in the generation of RAOs, where:
        Ae = External added mass, Be = External damping, Ce = External stiffness
    The remaining part will then be grouped in the wave force coefficients as a "modification". 
    
    Args:
        folder (str): folder containing SIF file
        filename (str): SIF filename
        as_simo_string (bool, optional): Return the coefficients as a simo string for input into a system file. Defaults to True.
        alpha (_type_, optional): Apply a constant scaling factor. Defaults to [1.0]*6.
        imosym (int, optional): Symmetry integer (0: None, 1: Y=0 symmetry, 2: X=0 and Y=0 symmetry). Defaults to 1.
        as_kN (bool, optional): Return as kN. Defaults to True.
        blend (list, optional): Array relating to the 6-DOF. If entry is 1, for that DOF returns coefficients affected by external A, B, C matrices. Defaults to [1,1,1,1,1,1].
        Be (np.ndarray, optional): External 6x6 damping matrix. Defaults to np.array((6,6)).
        Ae (np.ndarray, optional): External 6x6 added mass matrix. Defaults to np.array((6,6)).
        Ce (np.ndarray, optional): External 6x6 stiffness matrix. Defaults to np.array((6,6)).

    Returns:
        Tuple of np.ndarrays: angular frequencies, directions, forces
    """
    
    # experimental hs tp addition
    #import wafo.spectrum.models as wsm
    #S = wsm.Torsethaugen(hs, tp)
    
    sif, filename_tmp = open_sif(folder, filename)
    omegas = sif.get_angular_freqs()
    directions = sif.get_directions()
    X = sif.get_motion_raos()
    Fd = sif.get_excitationforce_raos()
    M = sif.get_bodymass()
    A = sif.get_addedmass()
    B = sif.get_potentialdamping()
    C = sif.get_hydrostatic_restoring()
    #Bv = sif.get_viscousdamping()    
    sif.close()
    os.remove(filename_tmp)
        
    F = np.zeros(Fd.shape, dtype=np.complex)
    alpha = np.array(alpha)
    blend = np.array(blend)
    
    for idir in range(directions.size):
        
        # experimental (perform crude integration to test)
        #H = X[2, idir, ::-1]
        #w = omegas[::-1]
        #Ve = np.sqrt(8/np.pi*np.trapz(S(w)*np.abs(H)**2,w))
        #Bv[2,2] = 0.125*1025*(np.pi*D**2)*Ve
        
        for iw, w in enumerate(omegas):
            lhs_matrix = -w**2*(M+A[:,:,iw]+Ae)+1j*w*(B[:,:,iw]+Be)+(C+Ce)
            raos = X[:, idir, iw]
            F[:,idir,iw] = lhs_matrix@raos*alpha*blend + Fd[:,idir,iw]*alpha*(1-blend)
    
    omegas = omegas[::-1]
    F = F[:,:,::-1]
    
    if as_kN: F = F/1000.
    
    if as_simo_string:
        string = []
        string.append("FIRST ORDER WAVE FORCE TRANSFER FUNCTION\n\n")
        string.append(f"{directions.size} {omegas.size} {imosym} 2")
        string.append("WAVE DIRECTIONS FORCE TRANSFER FUNCTION")
        string.append("\n".join([f"{idir+1:2d}  {d:.1f}" for idir, d in enumerate(directions)]))
        string.append("WAVE FREQUENCIES FORCE TRANSFER FUNCTION")
        string.append("\n".join([f"{iw+1:2d}  {w:.10e}" for iw, w in enumerate(omegas)]))
        names = "SURGE SWAY HEAVE ROLL PITCH YAW".split()
        for idof, name in enumerate(names):
            string.append(f"{name} FORCE TRANSFER FUNCTION")
            string.append('\n'.join([f"{idir+1:d}  {iw+1:d}  {np.abs(F[idof,idir,iw]):.10e} {np.angle(F[idof,idir,iw])*180./np.pi:.10e}" 
                            for idir in range(directions.size) for iw in range(omegas.size)]))
        return "\n".join(string)
    else:
        return omegas, directions, F        
    
def export_wamit_raos(sif_folder:str, sif_filename:str, wamit_filename:str="wamit.rao", 
                      wamit_periods=None, wamit_headings=None):
    sif, filename_tmp = open_sif(sif_folder, sif_filename)
    X = sif.get_motion_raos()
    periods = sif.get_periods()
    print(periods)
    headings = sif.get_directions()
    raos = sif.get_motion_raos()
    if wamit_periods is None: wamit_periods = periods
    if wamit_headings is None: wamit_headings = headings
    sif.close()
    os.remove(filename_tmp)

    # Modify the version to allow interpolation
    with open(wamit_filename, 'w') as fout:
        fout.write("external raos extracted from %s\n" % sif_filename)
        for per in wamit_periods:
            # TODO: interpolate for headings
            for iheading, head in enumerate(wamit_headings):
                for idof in range(6):
                    r = np.interp(per, periods, raos[idof, iheading])
                    fout.write("%.6e  %.6e  %d  %14.6e %14.6e %14.6e %14.6e\n"  % (per, head, idof + 1, 
                        np.abs(r), np.angle(r)*180./np.pi, r.real, r.imag))    


import numpy as np

def jonswap(f:np.ndarray, Hm0:float, Tp:float, gamma:float=3.3, sigma_low:float=.07, sigma_high:float=.09,
            g:float=9.81, method:str='yamaguchi', normalize:bool=False):
    '''Generate JONSWAP spectrum using a function
    taken from https://github.com/openearth/oceanwaves-python/blob/master/oceanwaves/spectral.py
    and then simplified (temporary solution to circumvent the need to use WAFO). Checked against
    wafo for specified gamma
    
    TODO: Set gamma's default value to one based on a formula
    
    To use:
    >> from teklib.wave_theory.spectrums import jonswap
    >> import matplotlib.pyplot as plt
    >> plt.plot(f, jonswap(f, 10, 15))
    >> plt.show()
    
    Parameters
    ----------
    f : numpy.ndarray
        Array of frequencies [Hz]
    Hm0 : float, numpy.ndarray
        Required zeroth order moment wave height
    Tp : float, numpy.ndarray
        Required peak wave period
    gamma : float
        JONSWAP peak-enhancement factor (default: 3.3)
    sigma_low : float
        Sigma value for frequencies <= ``1/Tp`` (default: 0.07)
    sigma_high : float
        Sigma value for frequencies > ``1/Tp`` (default: 0.09)
    g : float
        Gravitational constant (default: 9.81)
    method : str
        Method to compute alpha (default: yamaguchi)
    normalize : bool
        Normalize resulting spectrum to match ``Hm0``
    Returns
    -------
    E : numpy.ndarray
        Array of shape ``f, Hm0.shape`` with wave energy densities
    '''

    # Pierson-Moskowitz
    if method.lower() == 'yamaguchi':
        alpha = 1. / (.06533 * gamma ** .8015 + .13467) / 16.
    elif method.lower() == 'goda':
        alpha = 1. / (.23 + .03 * gamma - .185 / (1.9 + gamma)) / 16.
    else:
        raise ValueError('Unknown method: %s' % method)

    E_pm = alpha * Hm0**2 * Tp**-4 * f**-5 * np.exp(-1.25 * (Tp * f)**-4)

    # JONSWAP
    sigma = np.ones(f.shape) * sigma_low
    sigma[f > 1./Tp] = sigma_high

    E_js = E_pm * gamma**np.exp(-0.5 * (Tp * f - 1)**2. / sigma**2.)

    if normalize:
        # axis=0 seems to work fine with all kinds of inputs
        E_js *= Hm0**2. / (16. * np.trapz(E_js, f, axis=0))

    return E_js


from typing import Iterable, Tuple
import pydantic

import numpy as np
import matplotlib.pyplot as plt
from ..signal_analysis.specpack import specdata
from ..signal_analysis.timetools import find_peaks
from ..signal_analysis.stats import empirical_cdf as ecdf


class IrregularWave():
    def __init__(self,z:Iterable[float],t:Iterable[float],d:float=1E10,g:float=9.80665):
        """Instantiates a irregular wave class based on an inputted

        Args:
            z (Iterable[float]): time series of wave elevation
            t (Iterable[float]): time
            d (float, optional): water depth. Defaults to 1E10 m (deep water)
            g (float, optional): gravity. Defaults to 9.80665 m/s/s
        """
        m0,m1,m2,m4,T1,T2,Tp,Hm0,Sf,f = specdata(z,t)  
        self.m0 = m0
        self.m1 = m1
        self.m2 = m2
        self.m4 = m4
        self.T1 = T1
        self.T2 = T2
        self.Tp = Tp
        self.Hm0 = Hm0
        self.Sf = Sf
        self.f = f
        self.z = z
        self.t = t
        self.g = g
        self.d = d
        self.sp = 2*np.pi/g*Hm0/Tp**2 # steepness parameter

    def wave_crest_distribution(self, distribution="forristall2d"):
        """Returns a probability of exceedence function f(x) for the wave crest. This
        is 1 - F(x) where F(x) is the cumulative distribution function.
        
        Note that these are ensemble distributions, so comparing against a single
        wave crest distribution may cause misinterpretation. An exception is the
        HuangPDSR distribution, which provides a mean, upper 99% and lower 99% 
        curve for the probability distribution of a single realization. 

        Args:
            distribution (str, optional): The distribution function. Options include:
            
            "forristall2d": Function given in reference /1/ for a long crested wave
            "forristall3d": Function given in reference /1/ for a short crested wave
            "rayleigh": Rayleigh distributed wave heights (ref /1/). 
            Defaults to "forristall2d".            

            References:
            
            /1/ Forristall G.Z. Wave Crest Distribution: Observations and Second-Order Theory,
            American Meteorological Society, 1 August 2000.

        Returns:
            hc(x): crest height probability of exceedence function
        """
        from .regular import StokesSecond
        s1 = 2*np.pi/self.g*self.Hm0/self.T1**2
        ss = StokesSecond(height=self.Hm0, period=self.T1, depth=self.d)
        if distribution == "forristall2d" or distribution == "forristall3d":
            k1 = ss.wavenumber # independent of StokesSecond's height
            urs = self.Hm0/k1**2/self.d**3
            if distribution == "forristall2d":
                ac = 0.3536 + 0.2892*s1 + 0.1060*urs
                bc = 2 - 2.1597*s1 + 0.0968*urs**2
            else:
                ac = 0.3536 + 0.2568*s1 + 0.0800*urs
                bc = 2 - 1.7912*s1 - 0.5302*urs + 0.2824*urs**2                
            return lambda x, ac=ac, bc=bc: np.exp( -(x/ac/self.Hm0)**bc  ) #NOTE: defaults bind at creation
        elif distribution == "rayleigh":
            return lambda x: np.exp( -8*(x/self.Hm0)**2 )
        elif distribution == "huangPDSR" or distribution == "huangPDER":
            k1 = 4*np.pi**2/self.g/self.T1**2  # Deep water dispersion eq. used by Huang
            urs = self.Hm0/k1**2/self.d**3
            pa = np.r_[1,s1,s1**2,s1**3,s1**4,urs,urs**2]
            pb = np.r_[1,s1,s1**2,s1**3,s1**4,urs**2]
            if distribution == "huangPDSR":
                # Mean, lower, upper curves for probability < 1E-2)
                a1 = np.r_[0.2894, 12.3011, -662.6320, 12153.3466, -68031.8045, 0.3779, -3.7904]
                b1 = np.r_[1.5277, 67.2118, -3683.1338, 63759.0846, -336712.3631, -8.1382]
                a2 = np.r_[0.3712,1.0087,-43.0667,567.5292,-1173.1204,0.1276,0.3115]
                b2 = np.r_[2.006,4.841,-321.181,846.332,27223.189,0.832]
                ac1, bc1 = sum(a1*pa), sum(b1*pb)
                ac2, bc2 = sum(a2*pa), sum(b2*pb)
                xlimit = (-np.log(0.01))**(1/bc1)*self.Hm0*ac1
                mean = lambda x, ac1=ac1, ac2=ac2, bc1=bc1, bc2=bc2, xlimit=xlimit: \
                    np.where(x > xlimit, np.exp( -(x/ac1/self.Hm0)**bc1 ), np.exp( -(x/ac2/self.Hm0)**bc2) )             
                a1 = np.r_[0.1334, 13.0432, -751.0935, 14727.6571, -87711.0810, 0.0840, -5.2045]    
                b1 = np.r_[0.7965, 44.9229, -2525.6956, 47184.1324, -270084.1075, -13.3160]     
                a2 = np.r_[0.3516,2.3892,-105.4167,1557.3914,-5807.4512,0.0791,0.3550]
                b2 = np.r_[1.620,19.720,-909.665,10465.556,-22952.443,-3.726]
                ac1, bc1 = sum(a1*pa), sum(b1*pb)
                ac2, bc2 = sum(a2*pa), sum(b2*pb)
                xlimit = (-np.log(0.01))**(1/bc1)*self.Hm0*ac1
                upper = lambda x, ac1=ac1, ac2=ac2, bc1=bc1, bc2=bc2, xlimit=xlimit: \
                    np.where(x > xlimit, np.exp( -(x/ac1/self.Hm0)**bc1 ), np.exp( -(x/ac2/self.Hm0)**bc2) )             
                a1 = np.r_[0.3752, 9.1269, -446.8393, 7837.5720, -42682.1177, 0.8351, -5.7942,]
                b1 = np.r_[2.4637, 82.5688, -4300.5362, 68857.4691, -337914.7489, 2.5661]
                a2 = np.r_[0.3708,0.9835,-39.9404,598.3819,-2144.4920,0.2427,-0.5379]
                b2 = np.r_[2.212,10.951,-637.673,6707.056,-10244.141,4.505]
                ac1, bc1 = sum(a1*pa), sum(b1*pb)
                ac2, bc2 = sum(a2*pa), sum(b2*pb)
                xlimit = (-np.log(0.01))**(1/bc1)*self.Hm0*ac1
                lower = lambda x, ac1=ac1, ac2=ac2, bc1=bc1, bc2=bc2, xlimit=xlimit: \
                    np.where(x > xlimit, np.exp( -(x/ac1/self.Hm0)**bc1 ), np.exp( -(x/ac2/self.Hm0)**bc2) )             
                return (mean,lower,upper)
            elif distribution == "huangPDER":
                a1 = np.r_[0.3242, 11.7467, -652.1470, 12308.3001, -70529.2504, 0.3785, -3.8837]
                b1 = np.r_[1.7321,72.7179,-4093.6916,72132.2957,-386504.7502,-9.9594]
                a2 = np.r_[0.3733, 0.9398, -40.0095, 512.0601, -849.0734, 0.1294, 0.2882]
                b2 = np.r_[2.0411, 3.6068,-272.2806,-58.9649,32786.2976,0.9003]
                ac1, bc1 = sum(a1*pa), sum(b1*pb)
                ac2, bc2 = sum(a2*pa), sum(b2*pb)
                xlimit = (-np.log(0.01))**(1/bc1)*self.Hm0*ac1
                return lambda x, ac1=ac1, ac2=ac2, bc1=bc1, bc2=bc2, xlimit=xlimit: \
                    np.where(x > xlimit, np.exp( -(x/ac1/self.Hm0)**bc1 ), np.exp( -(x/ac2/self.Hm0)**bc2) )                 
        else:
            raise NotImplementedError(f"The {distribution} distribution is not implemented")
            
    def wave_height_distribution(self, distribution="forristall"):
        """Returns a probability of exceedence function f(x) for the wave height. This
        is 1 - F(x) where F(x) is the cumulative distribution function.

        Args:
            distribution (str, optional): The distribution function. Options include:
            
            "forristall": Function given in reference /1/
            "rayleigh": Rayleigh distributed wave heights
            Defaults to "forristall".            

            References:
            
            /1/ Forristall G.Z. On the Statistical Distribution of Wave Heights in a Storm,
            Journal of Geophysical Research Atmospheres, January 1978.
            /2/ Longuet-Higgins, M.S., On the statistical distribution of heights of sea waves.
            Journal of Marine Research 11, 245–266. 1952

        Returns:
            hc(x): wave height probability of exceedence function
        """
        if distribution == "rayleigh":
            return lambda x: np.exp( -2*(x/self.Hm0)**2 )
        elif distribution == "forristall": # 1978
            return lambda x: np.exp( -2.263*(x/self.Hm0)**2.126 )        
        else:
            raise NotImplementedError(f"The {distribution} distribution is not implemented")

    def plot_wave_crest_distribution(self, Hm0:float=None, Tp:float=None, figname:str=None, figsize:Tuple[float]=(10,6)):
        fig, ax = plt.subplots(figsize=figsize)
        ax.set_yscale("log");ax.set_xlim(0,2);ax.set_ylim(1E-6,1);ax.grid(True, which="both")
        ax.set_title(f"Wave crest distribution: Measured Hs={self.Hm0:.1f}m Tp={self.Tp:.1f}s")
        # Plot empirical distribution
        ix = find_peaks(self.z, global_peaks=True)
        Fe, ze = ecdf(len(self.z[ix])), np.sort(self.z[ix])
        ax.scatter(ze/self.Hm0,1-Fe, label="Empirical", color='r')
        # Plot semi-empirical distributions
        h = np.arange(0,30)
        if self.d > 100:
            f_huangPDSR = self.wave_crest_distribution(distribution="huangPDSR") # single realization distro
            f_huangPDER = self.wave_crest_distribution(distribution="huangPDER")
        f_forristall2d = self.wave_crest_distribution(distribution="forristall2d")
        f_rayleigh = self.wave_crest_distribution(distribution="rayleigh")
        if self.d > 100:
            ax.plot(h/self.Hm0, f_huangPDSR[0](h), color="black",linewidth=2,label="Huang PDSR (mean)")
            ax.plot(h/self.Hm0, f_huangPDSR[1](h), color="black",linewidth=2,label="Huang PDSR (lower)", linestyle='--')
            ax.plot(h/self.Hm0, f_huangPDSR[2](h), color="black",linewidth=2, linestyle='--')
            ax.plot(h/self.Hm0, f_huangPDER(h), color="grey",linewidth=2,label="Huang PDER")
        ax.plot(h/self.Hm0, f_forristall2d(h), color="green",linewidth=2,label="Forristall")
        ax.plot(h/self.Hm0, f_rayleigh(h), color="blue",linewidth=2,label="Rayleigh") 
        ax.legend(loc=0)
        ax.set_xlabel("$\eta_c / H_{m0}$  [-]", fontsize=14)
        ax.set_ylabel("POE [-]", fontsize=14)
        if figname is not None: 
            plt.savefig(figname) 
            plt.close(fig)
        else: 
            plt.show()
            return (fig, ax)      
        
    def plot_wave_height_distribution(self, Hm0:float=None, Tp:float=None, figname:str=None, figsize:Tuple[float]=(10,6)):
        fig, ax = plt.subplots(figsize=figsize)
        ax.set_yscale("log");ax.set_xlim(0,2.5);ax.set_ylim(1E-6,1);ax.grid(True, which="both")
        ax.set_title(f"Wave height distribution: Measured Hs={self.Hm0:.1f}m Tp={self.Tp:.1f}s")
        # Plot empirical distribution
        ix_max, = find_peaks(self.z, global_peaks=True, maxima=True)
        ix_min, = find_peaks(self.z, global_peaks=True, maxima=False)
        k = 0 if ix_min[0] < ix_max[0] else 1
        zh = [self.z[j] - self.z[i] for i, j in zip(ix_min, ix_max[k:])]
        Fe, ze = ecdf(len(zh)), np.sort(zh)
        ax.scatter(ze/self.Hm0,1-Fe, label="Empirical", color='r')
        # Plot semi-empirical distributions
        h = np.linspace(0,2.5*self.Hm0)
        f_forristall = self.wave_height_distribution(distribution="forristall")
        f_rayleigh = self.wave_height_distribution(distribution="rayleigh")
        ax.plot(h/self.Hm0, f_forristall(h), color="green",linewidth=2,label="Forristall")
        ax.plot(h/self.Hm0, f_rayleigh(h), color="blue",linewidth=2,label="Rayleigh") 
        ax.legend(loc=0)
        ax.set_xlabel("$H / H_{m0}$  [-]", fontsize=14)
        ax.set_ylabel("POE [-]", fontsize=14)
        if figname is not None: 
            plt.savefig(figname) 
            plt.close(fig)
        else: 
            plt.show()
            return (fig, ax)     
        
            
    @classmethod
    def wave_crest_distribution(cls, Hm0:float, Tp:float, T1:float, depth:float, 
                                distribution="forristall2d"):
        """Returns a probability of exceedence function f(x) for the wave crest. This
        is 1 - F(x) where F(x) is the cumulative distribution function.
        
        Note that these are ensemble distributions, so comparing against a single
        wave crest distribution may cause misinterpretation. An exception is the
        HuangPDSR distribution, which provides a mean, upper 99% and lower 99% 
        curve for the probability distribution of a single realization. 

        Args:
            distribution (str, optional): The distribution function. Options include:
            
            "forristall2d": Function given in reference /1/ for a long crested wave
            "forristall3d": Function given in reference /1/ for a short crested wave
            "rayleigh": Rayleigh distributed wave heights (ref /1/). 
            Defaults to "forristall2d".            

            References:
            
            /1/ Forristall G.Z. Wave Crest Distribution: Observations and Second-Order Theory,
            American Meteorological Society, 1 August 2000.

        Returns:
            hc(x): crest height probability of exceedence function
        """
        from .regular import StokesSecond
        g=9.80665
        s1 = 2*np.pi/g*Hm0/T1**2
        ss = StokesSecond(height=Hm0, period=T1, depth=depth)
        if distribution == "forristall2d" or distribution == "forristall3d":
            k1 = ss.wavenumber # independent of StokesSecond's height
            urs = Hm0/k1**2/depth**3
            if distribution == "forristall2d":
                ac = 0.3536 + 0.2892*s1 + 0.1060*urs
                bc = 2 - 2.1597*s1 + 0.0968*urs**2
            else:
                ac = 0.3536 + 0.2568*s1 + 0.0800*urs
                bc = 2 - 1.7912*s1 - 0.5302*urs + 0.2824*urs**2                
            return lambda x, ac=ac, bc=bc: np.exp( -(x/ac/Hm0)**bc  ) #NOTE: defaults bind at creation
        elif distribution == "rayleigh":
            return lambda x: np.exp( -8*(x/Hm0)**2 )


import numpy as np
import matplotlib.pyplot as plt
from .timetools import find_peaks


def empirical_cdf(n:int, kind:str='mean'):
    """
    Empirical cumulative distribution function given a sample size.

    Parameters
    ----------
    n : int
        sample size
    kind : str, optional
        - 'mean':  ``i/(n+1)`` (aka. Weibull method)
        - 'median' ``(i-0.3)/(n+0.4)``
        - 'symmetrical': ``(i-0.5)/n``
        - 'beard': ``(i - 0.31)/(n + 0.38)`` (Jenkinson's/Beard's method)
        - 'gringorten': ``(i - 0.44)/(n + 0.12)`` (Gringorten's method)

    Returns
    -------
    array
        Empirical cumulative distribution function

    Notes
    -----
    Gumbel recommended the following quantile formulation ``Pi = i/(n+1)``.
    This formulation produces a symmetrical CDF in the sense that the same
    plotting positions will result from the data regardless of
    whether they are assembled in ascending or descending order.

    Jenkinson's/Beard's method is based on the "idea that a natural
    estimate for the plotting position is the median of its probability
    density distribution".

    A more sophisticated formulation ``Pi = (i-0.3)/(n+0.4)`` approximates the
    median of the distribution free estimate of the sample variate to about
    0.1% and, even for small values of `n`, produces parameter estimations
    comparable to the result obtained by maximum likelihood estimations
    (Bury, 1999, p43)

    The probability corresponding to the unbiased plotting position can be
    approximated by the Gringorten formula in the case of type 1 Extreme
    value distribution.

    References
    ----------
    1. `Plotting positions <http://en.wikipedia.org/wiki/Q%E2%80%93Q_plot>`_, About plotting positions
    """

    n = float(n)
    i = np.arange(n) + 1
    if kind == 'mean':
        f = i / (n + 1.)
    elif kind == 'median':
        f = (i - 0.3) / (n + 0.4)
    elif kind == 'symmetrical':
        f = (i - 0.5) / n
    elif kind == 'beard':
        f = (i - 0.31) / (n + 0.38)
    elif kind == 'gringorten':
        f = (i - 0.44) / (n + 0.12)
    else:
        raise ValueError("Distribution type must be either 'mean','median','symmetrical','beard','gringorten'")

    return f


def fit_Weibull(x:np.ndarray,t:np.ndarray,maxima:bool=True, t0:float=0, duration:float=3, 
                savefig:bool=False, figname:str="weibull.png", percentile:float=90.0):
    """fits a Weibull distribution to either the maxima of minima of a timeseries x(t)

    Args:
        x (np.ndarray): timeseries x(t)
        t (np.ndarray): time in seconds
        maxima (bool, optional): fit to maxima if True, and minima if False. Defaults to True.
        t0 (float, optional): start time for windowing x(t) in seconds. Defaults to 0.
        duration (float, optional): duration of event in hours. Defaults to 3.
        savefig (bool, optional): set True to save the figure. If False figure axes is returned. Defaults to False.
        figname (str, optional): name of figure is savefig=True. Defaults to "weibull.png".

    Returns:
        tuple: 
            <duration>-hour sample extreme,
            <duration>-hour estimated extreme (Weibull fit)
            <duration>-hour mpm extreme (should equate to estimate extreme)
            <duration>-hour expected extreme
            <duration>-hour 90th percentile extreme
            if savefig=False: figure axes
    """
    from qats.stats.weibull import Weibull
    from scipy.stats import gumbel_l, gumbel_r, weibull_max, weibull_min, rayleigh
    sign = 1 if maxima else -1
    x = x*sign # flip space if minima
    # Window the time series
    ix = (t>t0) # & (t<t0+duration*60*60)
    x, t = x[ix], t[ix]
    # helper function to deal with minima
    correct_space = lambda x, maxima: x if maxima else -x 
    # find peaks 
    ix = find_peaks(x, global_peaks=True)
    # fit a Weibull distribution. Use qats for now
    weib_qats = Weibull.fit(data=x[ix])
    # get the Weibull params
    loc, scale, shape = weib_qats.params  # maxima_space
    # determine simulation hours (not duration for extreme)
    simhours = (t[-1]-t[0]) / 3600
    # get the associated asymptotic gumbel params for the duration of interest
    n = (duration / simhours) * weib_qats.data.size
    gloc, gscale = weib_qats.gumbel_parameters(n) # maxima_space (correct space with gumbel_l)
    # convert to relevant scipy implements
    gumbel, weibull = gumbel_r, weibull_min
    weib = weibull(loc=loc, scale=scale, c=shape) # Distribution of population extremes
    evd = gumbel(loc=gloc, scale=gscale)  # Extreme Value Distribution
    # Compute statistics (working with maxima)
    evd_mpm = gloc
    #evd_mean = gloc + sign*gscale*0.5772156649
    evd_mean = gloc + gscale*0.5772156649
    evd_90 = evd.ppf(percentile/100)    
    # make plots
    fig, ax1 = plt.subplots() # make axes
    yspace = lambda probability:  np.log(-np.log(1-probability))
    xspace = lambda xvalue: np.log(abs(xvalue-loc))
    # Add data points
    F_p, x_p = empirical_cdf(len(x[ix])), np.sort(x[ix])  # data points
    ax1.plot(xspace(x_p), yspace(F_p), color='r', linewidth=2, label="Empirical (data)")
    # Add weibull fit to plot
    x_fit = np.linspace(weib.ppf(0.01), weib.ppf(0.9999), 100) # x-range of interest for labelling
    F_fit = weib.cdf(x_fit)
    ax1.plot(xspace(x_fit),yspace(F_fit), color='k', label="Weibull fit")
    # Set yticks and labels
    yticklabels = np.r_[0.1, 0.2, 0.5, 0.7, 0.9, 0.99, 0.999, 0.9999]
    yticks = yspace(yticklabels)
    ax1.set_yticks(yticks)    
    ax1.set_yticklabels(yticklabels)    
    # Set xticks and labels
    xticklabels = weib.ppf(yticklabels)
    xticks = xspace(xticklabels)
    ax1.set_xticks(xticks)    
    ax1.set_xticklabels("%.1f" % (sign*l) for l in xticklabels)    
    ax1.grid(True)
    ax1.set_xlim(xticks[0], xticks[-1])
    ax1.set_ylim(yticks[0], yticks[-1])
    # Plot the most probable extreme for the duration from the weibull
    p = 1 / weib_qats.data.size * simhours / duration   
    #p = 1 - p if not maxima else p
    ax1.scatter([xspace(x_p[-1])], [yspace(F_p[-1])], color='r', marker='o', label=f"Sample extr ({sign*x_p[-1]:.1f})")
    ax1.scatter([xspace(weib.ppf(1-p))], [yspace(1-p)], color='k', marker='o', label=f"Weibull est. extr ({sign*weib.ppf(1-p):.1f})")
    ax1.legend(loc=0)
    if savefig:
        plt.savefig(figname)
        plt.close(fig)    
        return (sign*x_p[-1], sign*weib.ppf(1-p), sign*evd_mpm, sign*evd_mean, sign*evd_90)
    else:
        plt.close(fig)
        return (sign*x_p[-1], sign*weib.ppf(1-p), sign*evd_mpm, sign*evd_mean, sign*evd_90), ax1